#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_almost_equal
from ppr.ga import GASolver, get_shortest_path_ga

data = [np.array([[0, 0], [0, 1]]),
        np.array([[1, -1], [1, 0], [1, 1]]),
        np.array([[0, 2], [2, 2]])]
data_size = [len(q) for q in data]

solver = GASolver(data)

class TestGASOlver():
  def test_get_path(self):
    c = np.array([0, 1, 0])
    actual = solver.get_path(c)
    desired = np.array([[0, 0], [1, 0], [0, 2]])
    assert_almost_equal(actual, desired)
  
  def test_fitness(self):
    c = np.array([0, 1, 0])
    actual = solver.fitness(c)
    desired = -4.0
    assert_almost_equal(actual, desired)
  
  def test_create_chromosome(self):
    ch = solver.create_chromosome()
    
    # test path length
    actual1 = len(ch)
    desired1 = len(data)
    assert actual1 == desired1
    
    # test data type
    actual2 = [type(el) == np.int64 for el in ch]
    assert np.all(actual2)
    
    # test data range
    actual3 = [ch[i] in range(data_size[i]) for i in range(len(ch))]
    assert np.all(actual2)
  
  def test_create_population(self):
    # default size value
    actual1 = solver.create_population()
    s, n = actual1.shape
    assert s == 30
    assert n == len(data)
    # custom size value
    solver2 = GASolver(data, pop_size=15)
    actual1 = solver2.create_population()
    s, n = actual1.shape
    assert s == 15
    assert n == len(data)
  
  def test_mutate(self):
    # How do I test probabilistic functions?
    # There is a small change that the function works, but this test fails
    c = np.array([0, 1, 0])
    res = []
    for i in range(100):
      cm = solver.mutate(c, 0.8)
      res.append(np.all(c == cm))
    assert not np.all(res)
  
  def test_crossover(self):
    c1 = np.array([1, 2, 3, 4, 5])
    c2 = np.array([7, 8, 9, 10, 11])
    a1, a2 = solver.crossover(c1, c2, 1.0)
    assert not np.all(a1 == a2)
  
  def test_sort_population(self):
    pop = np.array([[0, 1, 0],
                    [0, 0, 0],
                    [1, 1, 1],
                    [1, 2, 1]])
    # f = [solver.fitness(c) for c in pop]
    # fitness of this pop: array([-4, -6, -5, -3])
    pop_sorted = np.array([[1, 2, 1],
                           [0, 1, 0],
                           [1, 1, 1],
                           [0, 0, 0]])
    actual = solver.sort_population(pop)
    assert_almost_equal(actual, pop_sorted)
  
  def test_new_generation(self):
    # for even populations the shape should be the same
    pop = np.array([[0, 1, 0],
                    [0, 0, 0],
                    [1, 1, 1],
                    [1, 2, 1]])
    actual1 = solver.new_generation(pop)
    assert_almost_equal(actual1.shape, pop.shape)
    # for add populations, size increases by one
    pop_odd = np.array([[0, 1, 0],
                        [0, 0, 0],
                        [1, 1, 1],
                        [1, 2, 1],
                        [1, 0, 1]])
    actual2 = solver.new_generation(pop_odd)
    a, _ = actual2.shape
    d, _ = pop_odd.shape
    assert a == d + 1
  
  def test_run(self):
    f_opt, p_opt, _ = solver.run()
    assert_almost_equal(f_opt, -3)
    # example data has two shortest paths with fitness -3
    assert np.all(p_opt == [1, 2, 0]) or np.all(p_opt == [1, 2, 1])

def test_get_shortest_path_ga():
  sol = get_shortest_path_ga(data)
  assert_almost_equal(sol['length'], 3)
  # example data has two shortest paths with fitness -3
  qs = sol['path']
  c1 = [1, 2, 0]; c2 = [1, 2, 1]
  d1 = np.vstack([data[i][c1[i]] for i in range(3)])
  d2 = np.vstack([data[i][c2[i]] for i in range(3)])
  assert np.allclose(qs, d1) or np.allclose(qs, d2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ppr.geometry import Rectangle

import numpy as np
from scipy.linalg import norm
from scipy.optimize import minimize
from numpy.testing import assert_almost_equal, assert_, assert_allclose

atol = 1e-15

def mimimum_distance(A1, b1, A2, b2):
    def obj(x):
        return np.sqrt(np.sum((x[:2] - x[2:])**2))
    cons = ({'type': 'ineq',
             'fun': lambda x: b1 - np.dot(A1, x[:2]),
             'jac': lambda x: np.hstack(( -A1, np.zeros((4, 2)) )) },
            {'type': 'ineq',
             'fun': lambda x: b2 - np.dot(A2, x[2:]),
             'jac': lambda x: np.hstack(( np.zeros((4, 2)), -A2 )) }
            )
    x_init = np.zeros(4)
    res = minimize(obj, x_init,
                   method='SLSQP',
                   constraints=cons)
    return obj(res['x'])

class TestRectanlge():
    def test_init_function(self):
        rec1 = Rectangle(0.5, 2, 1, 3, 0)
    
    def test_get_matrix_form(self):
        sq2 = np.sqrt(2)
        pi4 = np.pi / 4
        rects = [Rectangle(1, 2, sq2, 2*sq2, -np.pi / 4),
                 Rectangle(-5.5, -4.5, 2, 3,  np.pi / 2)]
        # expected results
        As = [np.array([[-1, -1], [1, -1], [1, 1], [-1, 1]]),
              np.array([[1, 0], [0, 1], [-1, 0], [0, -1]])]
        bs = [np.array([-3, 1, 7, 1]),
              np.array([-5.5, -2.5, 8.5, 4.5])]
        # normalize equations
        for i in range(len(As)):
            n = norm(As[i], axis=1)
            As[i] = As[i] / n[:, None]
            bs[i] = bs[i] / n
        for i in range(len(rects)):
            At, bt = rects[i].get_matrix_form()
            assert_allclose(At, As[i], atol=atol)
            assert_allclose(bt, bs[i], atol=atol)
    
    def test_get_matrix_form_in_optimisation_problem(self):
         rec1 = Rectangle(1, 2, 3, 2, 0)
         rec2 = Rectangle(-1, 5, 2, 4, np.pi/4)
         A1, b1 = rec1.get_matrix_form()
         A2, b2 = rec2.get_matrix_form()
         dist = mimimum_distance(A1, b1, A2, b2)
         assert_almost_equal(dist, 2.121320343529712)
    
    def test_check_self_collision(self):
        rec1 = Rectangle(1, 2, 3, 2, np.pi/3)
        actual = rec1.is_in_collision(rec1)
        desired = True
        assert_(actual == desired)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import matplotlib.pyplot as plt
from numpy.testing import assert_almost_equal
from ppr.geometry import Rectangle
from ppr.path import TolerancedNumber, TrajectoryPt
from ppr.path import cart_to_joint, get_shortest_path
from ppr.robot import Robot_3R, Robot_2P

class TestTolerancedNumber():
    def test_nominal_outside_bounds_error(self):
        with pytest.raises(ValueError) as info:
            a = TolerancedNumber(1.5, 0, 1)
        # check whether the error message is present
        msg = "nominal value must respect the bounds"
        assert(msg in str(info))

class TestTrajectoryPt():
    def test_plot(self):
        x = TolerancedNumber(1.5, 1.0, 2.0)
        y = TolerancedNumber(3.0, 2.9, 3.2)
        angle = TolerancedNumber(np.pi / 8, 0.0, np.pi / 4)
        tp = TrajectoryPt([x, y, angle])
        fig, ax = plt.subplots()
        plt.axis([0, 4, 0, 4])
        tp.plot(ax)

def test_cart_to_joint():
    robot2p = Robot_2P([1, 1])
    x = TolerancedNumber(2, 1, 3, samples=3)
    path = [TrajectoryPt([x, 1.0, 0]),
            TrajectoryPt([x, 1.5, 0]),
            TrajectoryPt([x, 2.0, 0])]
    actual = cart_to_joint(robot2p, path)
    desired = [np.array([[1, 1],
                         [2, 1],
                         [3, 1]]),
               np.array([[1, 1.5],
                         [2, 1.5],
                         [3, 1.5]]),
               np.array([[1, 2],
                         [2, 2],
                         [3, 2]])]
    assert(len(actual) == len(desired))
    for i in range(3):
        assert_almost_equal(actual[i], desired[i])

def test_cart_to_joint_no_scene():
    robot2p = Robot_2P([1, 1])
    path = []
    with pytest.raises(ValueError) as info:
        cart_to_joint(robot2p, path, check_collision=True)
    # check whether the error message is present
    msg = "scene is needed for collision checking"
    assert(msg in str(info))

def test_cart_to_joint_with_check_collision():
    robot2p = Robot_2P([1, 1])
    x = TolerancedNumber(2, 1, 3, samples=3)
    path = [TrajectoryPt([x, 1.0, 0]),
            TrajectoryPt([x, 1.5, 0]),
            TrajectoryPt([x, 2.0, 0])]
    scene = [Rectangle(0, 1.5, 1, 1, 0)]
    actual = cart_to_joint(robot2p, path, check_collision=True, scene=scene)
    desired = [np.array([[1, 1],
                         [2, 1],
                         [3, 1]]),
               np.array([[2, 1.5],
                         [3, 1.5]]),
               np.array([[2, 2],
                         [3, 2]])]
    assert(len(actual) == len(desired))
    for i in range(3):
        assert_almost_equal(actual[i], desired[i])

def test_get_shortest_path():
    Q = [np.array([[0, 0]]),
         np.array([[1, -1], [1, 0], [1, 1]]),
         np.array([[0, 2], [2, 2]])]
    res = get_shortest_path(Q)
    actual1 = res['success']
    actual2 = res['path']
    desired1 = True
    desired2 = [Q[0][0], Q[1][1], Q[2][0]]
    assert actual1 == desired1
    assert_almost_equal(actual2, desired2)

def test_get_shortest_path_failed():
    Q = [np.array([[0, 0]]),
         np.array([[1, -1], [1, 0], [1, 1]]),
         np.array([]),
         np.array([[0, 2], [2, 2]])]
    res = get_shortest_path(Q)
    assert res['success'] == False
    with pytest.raises(KeyError):
         res['path']
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from numpy.testing import assert_almost_equal, assert_
import pytest
from ppr.robot import Robot, Robot_3R, Robot_2P, Robot_2P3R

class TestRobot():
    def test_get_shapes(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        s = robot1.get_shapes([np.pi/2, -np.pi/2])
        desired = [np.array([[  0.0000000e+00,   0.0000000e+00],
                             [  1.2246468e-16,   2.0000000e+00],
                             [ -5.0000000e-02,   2.0000000e+00],
                             [ -5.0000000e-02,   3.0616170e-18]]),
                   np.array([[  1.2246468e-16,   2.0000000e+00],
                             [  1.0000000e+00,   2.0000000e+00],
                             [  1.0000000e+00,   2.0500000e+00],
                             [  1.2246468e-16,   2.0500000e+00]])]
        actual = s[0].get_vertices()
        assert_almost_equal(actual, desired[0])
        actual = s[1].get_vertices()
        assert_almost_equal(actual, desired[1])
    
    def test_wrong_joint_type(self):
        # check if exception is trown
        with pytest.raises(ValueError) as info:
            robot1 = Robot(['a', 'r'], [2, 1], [0, 0])
            robot1.fk_all_links([0, 0])
        # check if the message was present
        msg = "wrong joint type. Must be 'r' or 'p', not: a"
        assert(msg in str(info))
    
    def test_nonzero_base(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        robot1.base = [1, 2, np.pi/2]
        actual = robot1.fk_all_links([0, np.pi/2])
        desired = np.array([[ 1 ,  2 ,  np.pi/2],
                            [ 1 ,  4 ,  np.pi/2],
                            [ 0 ,  4 ,  np.pi  ]])

        assert_almost_equal(actual, desired)
    
    def test_plot(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(111)
        robot1.plot(ax1, [0.1, 1.5])
        #plt.show()
    
    def test_plot_kinematics(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        fig2 = plt.figure()
        ax2 = fig2.add_subplot(111)
        robot1.plot_kinematics(ax2, [0.1, 1.5])
        #plt.show()
    
    def test_plot_path(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(111)
        np.random.seed(42)
        path = np.random.randn(6, 2)
        robot1.plot_path(ax1, path)
        #plt.show()
    
    def test_plot_path_kinematics(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(111)
        np.random.seed(42)
        path = np.random.randn(6, 2)
        robot1.plot_path_kinematics(ax1, path)
        #plt.show()
    
    def test_set_joint_limits(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        robot1.set_joint_limits([(-1, 2), (-3, 2)])
    
    def test_set_link_inertia(self):
        robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        robot1.set_link_inertia([1, 1], [0.5, 0.3], [0.05, 0.03])

class TestRobot_3R():
    def test_init_function(self):
        robot3r = Robot_3R([1.5, 1.0, 1.0])
    
    def test_init_input_error(self):
        with pytest.raises(ValueError) as info:
            robot3r = Robot_3R([1.5, 1.0])
        msg = "This robot has 3 links, not: 2"
        assert(msg in str(info))
    
    def test_unreachable_inverse_kinematics(self):
        robot3r = Robot_3R([1.5, 1.0, 1.0])
        desired = {'success': False, 'info': "unreachable"}
        p_unreachable = np.array([99.0, 5.0, 3.0])
        actual = robot3r.ik(p_unreachable)
        assert_(actual == desired)
    
    def test_special_inverse_kinematics(self):
        robot3r = Robot_3R([1, 1, 1])
        p1 = np.array([1, 0, np.pi])
        ik_sol = robot3r.ik(p1)
        ik_desired = np.array([[0, 0, np.pi]])
        assert_almost_equal(ik_sol['q'], ik_desired)
    
    def test_random_inverse_kinematics(self):
        np.random.seed(42)
        q_test = np.random.rand(10, 3)
        q_test = q_test * 2 * np.pi - np.pi
        robot3r = Robot_3R([1.5, 1.0, 1.0])
        for qi in q_test:
            p = robot3r.fk(qi)
            ik_sol = robot3r.ik(p)
            q_sol = ik_sol['q']
            actual = [np.allclose(qj, qi) for qj in q_sol]
            assert_(np.any(actual))

class TestRobotDynamics():
    def test_fw_prop(self):
        pass

class TestRobot_2P():
    def test_init_function(self):
        robot2p = Robot_2P([1, 1])
    
    def test_init_input_error(self):
        with pytest.raises(ValueError) as info:
            robot3r = Robot_2P([1.5, 1.0, 1.0])
        msg = "This robot has 2 links, not: 3"
        assert(msg in str(info))
    
    def test_random_inverse_kinematics(self):
        np.random.seed(42)
        q_test = np.random.rand(10, 2)
        q_test = q_test * 2 * np.pi - np.pi
        robot2p = Robot_2P([1.5, 1.0])
        for qi in q_test:
            p = robot2p.fk(qi)
            ik_sol = robot2p.ik(p)
            q_sol = ik_sol['q']
            actual = [np.allclose(qj, qi) for qj in q_sol]
            assert_(np.any(actual))

class TestRobot_2P3R():
    def test_init_function(self):
        robot2p3r = Robot_2P3R([1.5, 1.0, 1.0, 0.5, 0.5])
    
    def test_init_input_error(self):
        with pytest.raises(ValueError) as info:
            robot2p3r = Robot_2P3R([1.5, 1.0])
        msg = "This robot has 5 links, not: 2"
        assert(msg in str(info))
    
    def test_unreachable_inverse_kinematics(self):
        robot2p3r = Robot_2P3R([1.5, 1.0, 1.0, 0.5, 0.5])
        desired = {'success': False, 'info': "unreachable"}
        p_unreachable = np.array([99.0, 5.0, 3.0])
        actual = robot2p3r.ik(p_unreachable)
        assert_(actual == desired)
    
    def test_random_inverse_kinematics(self):
        np.random.seed(42)
        q_test = np.random.rand(10, 5)
        q_test = q_test * 2 * np.pi - np.pi
        q_test[:, :2] = 0
        robot = Robot_2P3R([1.5, 1.0, 1.0, 0.5, 0.5])
        for qi in q_test:
            p = robot.fk(qi)
            ik_sol = robot.ik(p)
            q_sol = ik_sol['q']
            actual = [np.allclose(qj, qi) for qj in q_sol]
            #assert_almost_equal(actual, [True, True])
            assert_(np.any(actual))
    
    def test_set_joint_limits(self):
        robot2p3r = Robot_2P3R([1.5, 1.0, 1.0, 0.5, 0.5])
        robot2p3r.set_joint_limits([(0, 5), (0, 5), (), (), ()])
        desired = {'success': False, 'info': "unreachable"}
        pose = np.array([3, 3, 3.0])
        actual = robot2p3r.ik(pose)
        actual_q = actual['q']
        # there are 5 joint solutions expected
        assert_(len(actual_q), 5)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Function to define and proccess robot end-effector paths.
"""

import numpy as np
from matplotlib.patches import Wedge
from ppr.cpp.graph import Graph

class TolerancedNumber:
    """ A range on the numner line used to define path constraints
    
    It also has a nominal value in the range which can be used in cost
    functions of some sort in the future. For example, if it is preffered
    that the robot end-effector stays close to an optimal pose, but can
    deviate if needed to avoid collision.
    
    Attributes
    ----------
    n : float
        Nominal / preffered value for this number
    u : float
        Upper limit.
    l : float
        lower limit
    s : int
        Number of samples used to produce descrete version of this number.
    range : numpy.ndarray of float
        A sampled version of the range on the number line, including limits.
        The nominal value is not necessary included.
    
    Notes
    -----
    Sampling for orientation is done uniformly at the moment.
    In 3D this is no good and special sampling techniques for angles should be
    used.
    The sampled range is now an attribute, but memory wise it would be better
    if it is a method. Then it is only descritized when needed.
    But it the number of path points will probably be limited so I preffer this
    simpler implementation for now.
    """
    def __init__(self, nominal, lower_bound, upper_bound, samples=10):
        """ Create a toleranced number
        
        Nominal does not have to be in the middle,
        it is the preffered value when we ever calculate some kind of cost.
        
        Parameters
        ----------
        nominal : float
            Nominal / preffered value for this number
        lower_bound : float
        upper_bound : float
        samples : int
            The number of samples taken when a sampled version of the number
            is returned in the range attribute. (default = 10)
        
        Examples
        --------
        >>> a = TolerancedNumber(0.7, 0.0, 1.0, samples = 6)
        >>> a.range
        array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ])
        """
        if nominal < lower_bound or nominal > upper_bound:
            raise ValueError("nominal value must respect the bounds")
        self.n = nominal
        self.u = upper_bound
        self.l = lower_bound
        self.s = samples
        self.range = np.linspace(self.l, self.u, self.s)

class TrajectoryPt:
    """ Trajectory point for a desired end-effector pose in cartesian space
    
    This object bundles the constraints on the end-effector for one point
    of a path.
    
    Attributes
    ----------
    dim : int
        Pose dimensions, 3 for 2D planning, 6 for 3D planning.
    p : list of numpy.ndarray of float or ppr.path.TolerancedNumber
        Pose constraints for the end-effector (x, y, orientation).
        Can be a fixed number (a float) or a TolerancedNumber
    hasTolerance : list of bool
        Indicates which elements of the pose are toleranced (True) and
        fixed (False).
    p_nominal : list of float
        Same as p for fixed poses, the nominal value for a TolerancedNumber.
    timing : float
        Time in seconds it should be executed relative to the previous path
        point. Not used in current version.
    
    Examples
    --------
    Create a trajectory point at position (1.5, 3.1) with a symmetric
    tolerance of 0.4 on the x position.
    The robot orientation should be between 0 and pi / 4.
    (No preffered orientation, so assumed in the middle, pi / 8)

    >>> x = TolerancedNumber(1.5, 1.0, 2.0)
    >>> y = 3.1
    >>> angle = TolerancedNumber(np.pi / 8, 0.0, np.pi / 4)
    >>> tp = TrajectoryPt([x, y, angle])
    >>> tp.p_nominal
    [1.5, 3.1, 0.39269908169872414]
    
    A path is created by putting several trajectory points in a list.
    For example a vertical path with tolerance along the x-axis:
    
    >>> path = []
    >>> path.append(TrajectoryPt([TolerancedNumber(1.5, 1.0, 2.0), 0.0, 0]))
    >>> path.append(TrajectoryPt([TolerancedNumber(1.5, 1.0, 2.0), 0.5, 0]))
    >>> path.append(TrajectoryPt([TolerancedNumber(1.5, 1.0, 2.0), 1.0, 0]))
    >>> for p in path: print(p)
    [1.5, 0.0, 0]
    [1.5, 0.5, 0]
    [1.5, 1.0, 0]
    """
    def __init__(self, pee):
        """ Create a trajectory point from a given pose
        
        [x_position, y_position, angle last joint with x axis]
        
        Parameters
        ----------
        pee : list or numpy.ndarray of float or ppr.path.TolerancedNumber
            Desired pose of the end-effector for this path point,
            every value can be either a float or a TolerancedNumber
        """
        self.dim = len(pee)
        self.p = pee
        self.hasTolerance = [isinstance(pee[i], TolerancedNumber) for i in range(self.dim)]
        self.p_nominal = []
        for i in range(self.dim):
            if self.hasTolerance[i]:
                self.p_nominal.append(self.p[i].n)
            else:
                self.p_nominal.append(self.p[i])
        self.timing = 0.1 # with respect to previous point
    
    def __str__(self):
        """ Returns string representation for printing
        
        Returns
        -------
        string
            List with nominal values for x, y and orientation.
        """
        return str(self.p_nominal)

    def plot(self, axes_handle, show_tolerance=True):
        """ Visualize the path on given axes
        
        Parameters
        ----------
        axes_handle : matplotlib.axes.Axes
        show_tolerance : bool
            If True, the range for a TolerancedNumber is showed.
            A bar for x or y position, A wedge for orientation tolerance.
            (default True)
        """
        pn = self.p_nominal
        axes_handle.plot(pn[0], pn[1], 'k*')
        if show_tolerance:
            if self.hasTolerance[0]:
                do = -self.p[0].l + pn[0]
                du =  self.p[0].u - pn[0]
                axes_handle.errorbar(pn[0], pn[1], xerr=[[do], [du]], color=(0.5, 0.5, 0.5))
            if self.hasTolerance[1]:
                do = -self.p[1].l + pn[1]
                du =  self.p[1].u - pn[1]
                axes_handle.errorbar(pn[0], pn[1], yerr=[[do], [du]], color=(0.5, 0.5, 0.5))
            if self.hasTolerance[2]:
                # scale radius relative to trajectory point position
                radius = (pn[0] + pn[1]) / 20
                do = self.p[2].l * 180 / np.pi
                du = self.p[2].u * 180 / np.pi
                arc = Wedge((pn[0], pn[1]), radius, do, du, facecolor=(0.5, 0.5, 0.5, 0.5))
                axes_handle.add_patch(arc)

def create_grid(r):
    """ Create an N dimensional grid from N arrays
    
    Based on N lists of numbers we create an N dimensional grid containing
    all possible combinations of the numbers in the different lists.
    An array can also be a single float if their is now tolerance range.
    
    Parameters
    ----------
    r : list of numpy.ndarray of float
        A list containing numpy vectors (1D arrays) representing a sampled
        version of a range along an axis.
    
    Returns
    -------
    numpy.ndarray
        Array with shape (M, N) where N is the number of input arrays and
        M the number of different combinations of the data in the input arrays.
    
    Examples
    --------
    >>> a = [np.array([0, 1]), np.array([1, 2, 3]), 99]
    >>> create_grid(a)
    array([[ 0,  1, 99],
           [ 1,  1, 99],
           [ 0,  2, 99],
           [ 1,  2, 99],
           [ 0,  3, 99],
           [ 1,  3, 99]])
    """
    grid = np.meshgrid(*r)
    grid = [ grid[i].flatten() for i in range(len(r)) ]
    grid = np.array(grid).T
    return grid

def discretise(pt):
    """ Returns a discrete version of the range of a trajectory point
    
    Based on the sampled range in the Toleranced Numbers, a 3 dimensional grid
    representing end-effector poses that obey the trajectory point constraints.
    
    Parameters
    ----------
    pt : ppr.path.TrajectoryPt
    
    Returns
    -------
    numpy.ndarray
        Array with shape (M, 3) containing M possible poses for the robot
        end-effector that  obey the trajectory point constraints.
    
    Examples
    --------
    >>> x = TolerancedNumber(1, 0.5, 1.5, samples=3)
    >>> y = TolerancedNumber(0, -1, 1, samples=2)
    >>> pt = TrajectoryPt([x, y, 0])
    >>> discretise(pt)
    array([[ 0.5, -1. ,  0. ],
           [ 1. , -1. ,  0. ],
           [ 1.5, -1. ,  0. ],
           [ 0.5,  1. ,  0. ],
           [ 1. ,  1. ,  0. ],
           [ 1.5,  1. ,  0. ]])
    """
    r = []
    for i in range(pt.dim):
        if pt.hasTolerance[i]:
            r.append(pt.p[i].range)
        else:
            r.append(pt.p[i])
    grid = create_grid(r)
    return grid

def cart_to_joint(robot, traj_points, check_collision = False, scene=None):
    """ Convert a path to joint space by descretising and ik.
    
    Every trajectory point in the path is descretised, then for all these
    poses the inverse kinematics are solved.
    
    Parameters
    ----------
    robot : ppr.robot.Robot
    traj_points : list of ppr.path.TrajectoryPt
    check_collision : bool
        If True, a joint solution is only accepted if it does not collide
        with the objects in the scene. (default false)
        Self collision is not checked but assumed to be ensured by the joint
        limits.
    scene : list of ppr.geometry.Rectangle
        A list of objects with which the robot could collide.
    
    Returns
    -------
    list of numpy.ndarray of floats
        A list of arrays with shape (M, ndof) representing possible joint
        positions for every trajectory point.
        The arrays in this list could be very big!
    """
    # input validation
    if check_collision:
        if scene == None:
            raise ValueError("scene is needed for collision checking")
    
    # get discrete version of trajectory points
    cart_traj = []
    for pt in traj_points:
        cart_traj.append(discretise(pt))

    # solve inverse kinematics for every samples traj point
    # I could add some print statements to have info on unreachable points
    joint_traj = []
    for cart_vec in cart_traj:
        qi = []
        for cart_pt in cart_vec:
            sol = robot.ik(cart_pt)
            if sol['success']:
                for qsol in sol['q']:
                    if check_collision:
                        if not robot.check_collision(qsol, scene):
                            qi.append(qsol)
                    else:
                        qi.append(qsol)
        joint_traj.append(np.array(qi))
    return joint_traj

def get_shortest_path(Q):
    """ Calculate the shortest path from joint space data
    
    When the path with trajectory points is converted to joint space,
    this data can be used to construct a graph and look for the shortest path.
    The current distance metrix is the l1-norm of joint position difference
    between two points.
    
    I still have to implement maximum joint movement and acceleration limits.
    So technically this will always find a shortest path for now.
    
    Parameters
    ----------
    Q : list of nympy.ndarrays of float
        A list with the possible joint positions for every trajectory point
        along a path.
    
    Returns
    -------
    dict
        A dictionary with a key 'success' to indicate whether a path was found.
        If success is True, then the key 'path' contains a list with the joint
        position for every trajectory point that gives the shortest path.
    
    Notes
    -----
    I have a problem with swig type conversions. Therefore the type of the
    input data is checked and changed from float64 to float32.
    """
    Q = _check_dtype(Q)

    n_path = len(Q)
    # initialize graph
    g = Graph()
    for c in Q:
        if len(c) == 0:
            # one of the trajectory points is not reachable
            return {'success': False}
        g.add_data_column(c)
    g.init_dijkstra()

    # run shortest path algorithm
    g.run_dijkstra()

    # print result
    # g.print_graph()
    g.print_path()

    # get joint values for the shortest path
    p_i = g.get_path(n_path)
    print(p_i)

    if p_i[0] == -1:
        return {'success': False}
    else:
        res = []
        for k, i in zip(range(n_path), p_i):
            # TODO ugly all the "unsave" typecasting
            qki = Q[k][i].astype('float64')
            res.append(qki)
        
        return {'success': True, 'path': res}

def _check_dtype(Q):
    """ Change type if necessary to float32
    
    Due to an unresolved issue with swig and numpy, I have to convert the type.
    
    Parameters
    ----------
    Q : list of nympy.ndarrays of float
        A list with the possible joint positions for every trajectory point
        along a path.
    
    Returns
    -------
    list of nympy.ndarrays of float32
    """
    if Q[0].dtype != 'float32':
        print("converting type of Q")
        for i in range(len(Q)):
            Q[i] = Q[i].astype('float32')
    
    return Q



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ppr.optim import q_derivatives, collision_ieq_con
from numpy.testing import assert_almost_equal
import numpy as np
from ppr.robot import Robot_3R
from ppr.geometry import Rectangle

def test_q_derivatives():
    q = np.array([[0, 0],
                  [1, 2],
                  [3, 4],
                  [7, 2]])
    dq_desired = np.array([[10.,  20.],
                           [15.,  20.],
                           [30.,   0.],
                           [40., -20.]])
    ddq_desired = np.array([[ 50.,    0.],
                            [100., -100.],
                            [125., -200.],
                            [100., -200.]])
    dq, ddq = q_derivatives(q)
    assert_almost_equal(dq, dq_desired)
    assert_almost_equal(ddq, ddq_desired)

def test_q_derivatives_with_known_function():
    t = np.linspace(0, 10, 50)
    dt = t[1]-t[0]
    x   = np.array([ np.sin(t),  np.cos(t), t**3     ]).T
    dx  = np.array([ np.cos(t), -np.sin(t), 3 * t**2 ]).T
    ddx = np.array([-np.sin(t), -np.cos(t), 6 * t    ]).T
    dx_test, ddx_test = q_derivatives(x, dt=dt)
    
    # the derivatis are not that good at the moment
    np.testing.assert_allclose(dx_test[10:40, :], dx[10:40, :],
                               rtol=1, atol=0.01)
    np.testing.assert_allclose(ddx_test[10:40, :], ddx[10:40, :],
                               rtol=1, atol=0.01)

class TestConstraints():
    def test_collision_ieq_con(self):
        robot = Robot_3R([2.1, 3.0, 1.8])
        rects = [Rectangle(0, 0, 1, 1, 0),
                 Rectangle(5, 0, 1, 2, 0.3)]
        q_test = [np.array([np.pi/4, -np.pi/4, 0.0])]
        actual = collision_ieq_con(q_test, robot, rects)
        desired = np.array([-1.0353553,  1.7665554, 
                            0.4849242,  0.0384695, 
                            0.4849242, -1.0532455])
        assert_almost_equal(actual, desired)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some classes for "virtual" storage, where variables in a virtual array
are calculated on the fly.
Applied to path to joint transformation for 2P 3R robot
"""

import numpy as np
from ppr.path import TolerancedNumber

def cp(array):
  """ Cumulative product
  """
  return np.cumprod(array)[-1]

def setup_ranges(trajectory_pt):
  """ Get discrete axis arrays from trajectory point
  """
  ranges = []
  for i in range(trajectory_pt.dim):
    if trajectory_pt.hasTolerance[i]:
      ranges.append(trajectory_pt.p[i].range)
    else:
      ranges.append(np.array([trajectory_pt.p[i]]))
  return ranges

class VirtualArray():
  """ Virtual version of an N D array to store big grids.
  
  By supplying the discretised individual axes of an N-dimensional space
  this class can calculate grid points on the fly, instead of calculating
  and storing the whole N-dimensional array at once.
  
  In essence this converts an single integer number from the range
  [0, n1*n2*n3-1] into a list of values from the range arrays.
  [range1[n1], range[1][n2], ...] in a deterministic way.
  
  Attributs
  ---------
  ranges : list of numpy.ndarray
  dim : int
    The number of dimensions of the virtual array, the length of the list
    ranges.
  sizes : list of int
    The size of the different discrete dimensions.
  total_size : int
    The total number of grid points that can be accesed.
  """
  def __init__(self, ranges):
    """ Create virtual array
    
    Paramters
    ---------
    ranges : list of numpy.ndarray
      Every element in the list represents a discretised axis in and
      n-dimensional space, where n is the list length.
    """
    self.ranges = ranges
    self.dim = len(ranges)
    self.sizes = [len(a) for a in self.ranges]
    self.total_size  = cp(self.sizes)
  
  def __getitem__(self, index):
    """ Implement index slicing
    
    The method get's the syntac object[i] working.
    Therefore the object acts as a 1D numpy array.
    """
    si = self.sample(index, self.sizes)
    return [self.ranges[k][si[k]] for k in range(self.dim)]

  def sample(self, I, l):
    a = [0] * len(l)
    l = np.array(l)
    
    # initial case i = 0
    D, R = divmod(I, l[0])
    a[0] = int(R)
    I   -= R
    for i in range(1, len(l) - 1):
        D, R = divmod(I, cp(l[0:i+1]))
        a[i] = int(R / cp(l[0:i]))
        I -= R
    # last element case
    a[-1] = int(D)
    
    return a

class VirtualJointSolutions():
  """ Virtual container of all possible joint solutions
  
  Currently implemented for the 2P 3R robot
  Calculated on the fly.
  
  Main problem: joint configuration.
  Should I check it or can a joint configuration index result in a different
  configuration on different occasions?
  """
  def __init__(self, robot, trajectory_pt, scene=None):
    # path tolerances
    ranges = setup_ranges(trajectory_pt)
    # ik redundancy, use default joint limits
    q1 = np.linspace(0, 1, robot.ik_samples[0])
    q2 = np.linspace(0, 1, robot.ik_samples[1])
    ranges.append(q1)
    ranges.append(q2)
    # robot configurations
    ranges.append(np.array([0, 1]))
    self.ranges = ranges
    self.va = VirtualArray(ranges)
    self.size = self.va.total_size
    self.scene = scene
    self.robot = robot

  def __getitem__(self, index):
    data = self.va[index]
    pee = [data[0], data[1], data[2]] # end-effector pose
    qf  = [data[3], data[4]]          # fixed joint values
    jc  = int(data[5])                # joint configuration index
    sol = self.robot.ik_fixed_joints(pee, qf)
    if sol['success']:
      qs = sol['q']
      if jc == 0:
        if self.scene is not None:
          if self.robot.check_collision(qs[0], self.scene):
            return {'success': False}
        return {'success': True, 'q': qs[0]}
      elif jc == 1 and len(qs) > 1:
        if self.scene is not None:
          if self.robot.check_collision(qs[1], self.scene):
            return {'success': False}
        return {'success': True, 'q': qs[1]}
      else:
        return {'success': False}
    else:
      return {'success': False}
  
  def __len__(self):
    return self.size



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Find global optimal path using a genetic algorithm
"""

import numpy as np
from numpy.random import randint, rand, choice

def get_shortest_path_ga(Q, *arg, **kwarg):
    # setup problem
    solver = GASolver(Q, *arg, **kwarg)
    
    # Check whether at least one path exists
    if 0 in solver.Q_size:
        return {'success': False}
    
    # run ga algorithm
    f_opt, path_index, fvec = solver.run()
    path = [ Q[i][path_index[i], :] for i in range(solver.n_path) ]
    return {'success': True, 'path': path, 'length': -f_opt}

#=============================================================================
# GA class
#=============================================================================

class GASolver():
  """ Find a shortest path in a discretized joint solution space Q using a
  genetic algorithm.
  
  Attributes
  ----------
  Q : list of numpy.ndarray of float
  n_path : int
    Number of path points, len(Q).
  Q_size : list of int
    Number of joint solutions for every path point in Q.
  cross_rate : float
    Crossover rate.
  mut_rate : float
    Mutation rate.
  gen : int
    The number of generations that the genetic algorithm goas through.
  """
  def __init__(self, Q, cross_rate=0.8, mut_rate=0.1, pop_size=30, gen=100):
    self.Q = Q
    self.n_path = len(Q)
    self.Q_size = [len(q) for q in Q]
    
    self.cross_rate = cross_rate
    self.mut_rate = mut_rate
    self.pop_size = pop_size
    self.generations = gen
  
  def run(self):
    cnt = self.generations
    pop = self.create_population()
    f_opt = -np.inf
    f_all = []
    while(cnt > 0):
        pop = self.sort_population(pop)
        
        f_cur = self.fitness(pop[0])
        f_all.append(f_cur)
        if f_cur > f_opt:
            f_opt = f_cur
        
        pop = self.new_generation(pop)
        cnt -= 1
    f_all = np.array(f_all)
    return f_opt, pop[0], f_all
  
  def create_chromosome(self):
    """ Create array of integers representing a joint trajectory in Q.
    
    Returns
    -------
    numpy.ndarray of int
      Array with shape (n_path,) where every element represents the index of
      a joint solution for the corresponding path point in Q.
    """
    return np.array([randint(0, self.Q_size[i]) for i in range(self.n_path)])

  def create_population(self):
    """ Create a array with a chromosome on every row
    
    Parameters
    ----------
    size : int
      Population size, the number of chromosomes in the array.
    
    Returns
    -------
    numpy.ndarray of int
      Array with shape (size, n_path) containing a chromosome on every row.
    """
    pop = [self.create_chromosome() for i in range(self.pop_size)]
    return np.array(pop)
  
  def fitness(self, c):
    """ Evaluate the cost of a chromosome, currently joint motion is minimised.
    """
    path = self.get_path(c)
    dp   = np.diff(path, axis=0)
    return -np.sum(np.abs(dp))
  
  def mutate(self, c, rate):
    """ Mutate a chromosome
    
    A random value in the chromosome is changed when a uniform random number
    is lower than rate, the given mutation rate.
    
    TODO Fix low mutation rate if the random replacement gene is the same
    as the original. Only relevant if there is a small number of joint
    solutions for a path point.
    
    Parameters
    ----------
    numpy.ndarray of int
      Array with shape (n_path,) where every element represents the index of
      a joint solution for the corresponding path point in Q.
    rate : float
      A number between 0 and 1 representing the probability that a chromosome
      will be mutated.
    """
    if rand() < rate:
        c_m = c.copy()
        m_i = randint(0, len(c))
        c_m[m_i] = randint(0, self.Q_size[m_i])
        return c_m
    else:
        return c
    
  def crossover(self, ch1, ch2, rate):
    """ Single point crossover between two chromosomes
    
    With a given rate, crossover two chromosomes by swapping the content
    op to a given random index. For example:
    [1, 2, 3, 4] and [10, 11, 12, 13] => [10, 11, 12, 4] and [1, 2, 3, 13]
    
    Parameters
    ----------
    ch1 : numpy.ndarray of int
      The first chromosome
    ch2 : numpy.ndarray of int
      The second chromosome
    rate : float
      A number between 0 and 1 representing the probability that crossover
      will occur.
    
    Returns
    -------
    tuple of numpy.ndarray of int
      A tuple containing the two crossed over chromosomes.
    """
    if rand() < rate:
        index = randint(0, len(ch1))
        c1 = np.hstack(( ch1[:index], ch2[index:] ))
        c2 = np.hstack(( ch2[:index], ch1[index:] ))
    else:
        c1 = ch1
        c2 = ch2
    return c1, c2
  
  def sort_population(self, pop):
    """ Sort the population from highest fitness to lowest fitness.
    """
    fit = [self.fitness(ci) for ci in pop]
    # change sign because argsort default to sort from small to big
    # and we want the biggest fitness at the first.
    fit = -np.array(fit)
    return pop[ fit.argsort() ]
  
  def new_generation(self, pop):
    """ Create a new population based on the previous one.
    """
    p_size = len(pop) # if an odd number, the next pop will have size+1
    current_size = 0
    new_pop = []
    while(current_size < p_size):
        # select couple
        # better one of the first elemens in pop, because they are ordered
        # based on fitness
        prob = np.linspace(1, 0, p_size)
        prob = prob / np.sum(prob)
        prob = list(prob)
        couple = choice(p_size, 2, p = prob)
        # mate couple
        c1, c2 = self.crossover(pop[couple[0]],
                                pop[couple[1]],
                                self.cross_rate)
        # mutate children
        c1 = self.mutate(c1, self.mut_rate)
        c2 = self.mutate(c2, self.mut_rate)
        # add children to new population
        new_pop.append(c1)
        new_pop.append(c2)
        current_size += 2
    return np.array(new_pop)
  
  def get_path(self, c):
    """ Get joint space path from data Q based on list of indices
    
    Parameters
    ----------
    c : list of int
      List of the index of the joint solution that should be selected for
      every path point in Q.
    
    Returns
    -------
    numpy.ndarray
      Array of shape (n_path, ndof) containing a discrete path in joint space.
    """
    return np.array([self.Q[i][c[i]] for i in range(self.n_path)])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from .geometry import Rectangle, rotation
from .path import TolerancedNumber

class Robot:
    """ Base class for a serial manipulator
    
    Attributes
    ----------
    ndof : int
        Number of degrees of freedom, equal to the number of links.
    jt : list of char
        List of joint type for every joint, 'r' or 'p'.
    d : list of float
        Link length for links with revolute joint.
        Initial joint length when link has prismatic joint.
    a : Link angle relative to previous link for prismatic joint.
        Initial joint angle for links with revolute joint.
    lw : list of float
        Link width.
    ll : list of float
        Link length. In case of a prismatic joint, this is de initial length.
    adapt_ll : bool
        When their are prismatic joints, the length of those links can change
        accroding to the given joint values if set to True. (default is True)
    base : numpy.ndarray of float
        Pose of the robot base (x, y, angle), where the first joint is
        located.
    """
    
    def __init__(self, joint_type, j_offset, j_angle):
        """ Create general robot object
        
        Parameters
        ----------
        joint_type : list of char
            Use 'p' for prismatic and 'r' for revolute.
        j_offset : list of float
            Link length. Initial length for a prismatic joint.
        j_angle : list of float
            Link angle. Initial angle for revolute joint.
        """
        
        # all input lists must have the same length
        self.ndof = len(joint_type)
        self.jt = joint_type
        self.d = j_offset
        self.a = j_angle
        
        # default collision shapes
        self.lw = [0.05] * self.ndof
        self.ll = j_offset
        self.adapt_ll = True
        
        # default base pose
        self.base = np.array([0, 0, 0])
    
    def set_joint_limits(self, joint_limits):
      """ Set joint limits for inverse kinematics
      
      Joint limits are not in the default init function because it would
      become to clumsy and a lot can be done without them.
      
      joint_limits : list of tuple of float
        A list with for every link the lower and upper joint limit, given as
        a tuple of length 2. For example [(-1, 1), (0, 2)]
      """
      self.jl = joint_limits
    
    def set_link_inertia(self, mass, cg_position, Icg):
        """ Set mass, position of center of gravity and Inertia for links.
        
        Paramters
        ---------
        mass : list of float
            List with masses for every link.
        cg_position : list of float
            Position of center of gravety along link.
        Icg : list of float
            Mass moment of inertia of link around center of gravity.
        """
        self.m = mass
        self.c = cg_position
        self.I = Icg
    
    def fk(self, q):
        """ Calculate forward kinematics
        
        Uses the fk_all_links function because it's almost the same.
        
        Parameters
        ----------
        q : list or numpy.array of float
            Joint angles
        
        Examples
        --------
        
        Example for a simple 2R robot.
        
        >>> robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        >>> robot1.fk([np.pi/2, -np.pi/2])
        array([ 1.,  2.,  0.])
        
        """
        return self.fk_all_links(q)[-1]
    
    def fk_all_links(self, q):
        """ Calculate position of all links, and the end effector
        
        The end-effector is currently assumed at the end of the last link.
        Usefull for collision detection and visualizing the robot.
        
        Parameters
        ----------
        q : list or numpy.array of float
            Joint angles
        
        Returns
        -------
        numpy.ndarray
            Array of shape (ndof+1, 3) containing position and orientation
            (x, y, angle) of all links and the end-effector.
        
        Examples
        --------
        
        Example for a simple 2R robot.
        
        >>> robot1 = Robot(['r', 'r'], [2, 1], [0, 0])
        >>> robot1.fk_all_links([np.pi/2, -np.pi/2])
        array([[  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  1.22464680e-16,   2.00000000e+00,   1.57079633e+00],
               [  1.00000000e+00,   2.00000000e+00,   0.00000000e+00]])
    
        Example for cartesian x-y robot.
        
        >>> robot2 = Robot(['p', 'p'], [1, 1], [0, np.pi/2])
        >>> robot2.fk_all_links([0.3, 0.6])
        array([[ 0.        ,  0.        ,  0.        ],
               [ 0.3       ,  0.        ,  0.        ],
               [ 0.3       ,  0.6       ,  1.57079633]])
        """
        pos = np.zeros((self.ndof+1, 3))
        for i in range(self.ndof):
            if self.jt[i] == 'r':
                # absolute orientation link i
                pos[i+1, 2] = pos[i, 2] + q[i]
                # position
                pos[i+1, 0] = pos[i, 0] + self.d[i] * np.cos(pos[i+1, 2])
                pos[i+1, 1] = pos[i, 1] + self.d[i] * np.sin(pos[i+1, 2])
            elif self.jt[i] == 'p':
                # absolute orientation link i
                pos[i+1, 2] = pos[i, 2] + self.a[i]
                # position
                pos[i+1, 0] = pos[i, 0] + q[i] * np.cos(pos[i+1, 2])
                pos[i+1, 1] = pos[i, 1] + q[i] * np.sin(pos[i+1, 2])
            else:
                msg = "wrong joint type. Must be 'r' or 'p', not: "
                msg += self.jt[i]
                raise ValueError(msg)
        
        # transform from base to world
        if self.base[2] == 0:
            # only translation
            pos[:, :2] = pos[:, :2] + self.base[:2]
        else:
            pos[:, 2] += self.base[2]
            R = rotation(self.base[2])
            pos[:, :2] = np.dot(R, pos[:, :2].T).T + self.base[:2]
        return pos
    
    def get_shapes(self, q):
        """ Get a list with rectangles (shapes in the future?) for all
        robot links in position q
        
        Parameters
        ----------
        q : list or numpy.array of float
            Joint angles
        
        Returns
        -------
        list of ppr.geometry.Rectangle
            List of the link shapes.
        """
        p = self.fk_all_links(q)
        shapes = []
        for i in range(self.ndof):
            shapes.append(self.get_link_shape(i, p[i, 0], p[i, 1], p[i+1, 2], q[i]))
        return shapes

    def get_link_shape(self, i, xi, yi, phii, qi):
        """ Get the rectangle representing link i of the robot
        
        Parameters
        ----------
        i : int
            Link index from 0 to ndof.
        xi : float
            x-position of link origin.
        yi : float
            y-position of link origin.
        phii : float
            Absolute angle of link with x-axis.
        qi : float
            Joint value for the joint of this link, usefull for prismatic
            joints to show variable length.
        
        Returns
        -------
        ppr.geometry.Rectangle
            A rectangle (or in the future a shape) oject representing the
            robot link.
        
        Examples
        --------
        
        Illustrate variable vs fixed prismatic joint length.
        
        >>> robot2 = Robot(['p', 'p'], [2, 1], [0, np.pi/2])
        >>> link0 = robot2.get_link_shape(0, 0, 0, 0, 1.5)
        >>> link0.get_vertices()
        array([[ 0.  ,  0.  ],
               [ 1.5 ,  0.  ],
               [ 1.5 ,  0.05],
               [ 0.  ,  0.05]])
        >>> robot2.adapt_ll = False
        >>> link0 = robot2.get_link_shape(0, 0, 0, 0, 1.5)
        >>> link0.get_vertices()
        array([[ 0.  ,  0.  ],
               [ 2.  ,  0.  ],
               [ 2.  ,  0.05],
               [ 0.  ,  0.05]])
        """
        if self.adapt_ll and (self.jt[i] == 'p'):
            # link length = joint length for prismatic case
            return Rectangle(xi, yi, qi, self.lw[i], phii)
        else:
            # fixed link length
            return Rectangle(xi, yi, self.ll[i], self.lw[i], phii)
    
    def check_collision(self, q, scene):
        """ Check for collision between the robot other_rectangles
        
        Parameters
        ----------
        q : list or numpy.ndarray of float
            Joint position
        scene : list of ppr.geometry.Rectangle
            A list with collision shapes.
        
        Returns
        -------
        bool
            True if one of the robot links collides with one of the collision
            shapes. False otherwise.
        """
        for recti in self.get_shapes(q):
            for rectj in scene:
                if recti.is_in_collision(rectj):
                    return True # exit if a collision is found
        return False
    
    def plot_kinematics(self, axes_handle, q, *arg, **karg):
        """ Plot robot using lines and dots
        
        Ingnore the (rectanlge) collision shapes.
        
        Parameters
        ----------
        axes_handle : matplotlib.axes.Axes
            Axes handle to draw the robot on.
        q : np.array or list of floats
            Joint values
        """
        p = self.fk_all_links(q)
        axes_handle.plot(p[:, 0], p[:, 1], *arg, **karg)
    
    def plot_path_kinematics(self, axes_handle, qp):
        """ Plot robot for more than one joint position using lines and dots
        
        A color gradient is used to from darker to lighter.
        
        Parameters
        ----------
        axes_handle : matplotlib.axes.Axes
            Axes handle to draw all the robots
        qp : np.array of floats
            Array of shape (N, ndof) representing N different joint positions,
            ndof is the number of robot joints.
        """
        alpha = np.linspace(1, 0.2, len(qp))
        for i, qi in enumerate(qp):
            self.plot_kinematics(axes_handle, qi, color=(0.1, 0.2, 0.5, alpha[i]))

    def plot(self, axes_handle, q, *arg, **karg):
        """ Plot robot
        
        Plot robot shape in given joint position.
        
        Parameters
        ----------
        axes_handle : matplotlib.axes.Axes
            Axes handle to draw the robot on.
        q : np.array or list of floats
            Joint values
        """
        for recti in self.get_shapes(q):
            recti.plot(axes_handle, *arg, **karg)

    def plot_path(self, axes_handle, qp):
        """ Plot robot for more than one joint position
        
        A color gradient is used to from darker to lighter.
        
        Parameters
        ----------
        axes_handle : matplotlib.axes.Axes
            Axes handle to draw all the robots
        qp : np.array of floats
            Array of shape (N, ndof) representing N different joint positions,
            ndof is the number of robot joints.
        """
        # draw robot more transparant to the end of the path
        alpha = np.linspace(1, 0.2, len(qp))
        
        for i, qi in enumerate(qp):
            for rect in self.get_shapes(qi):
                rect.plot(axes_handle, color=(0.1, 0.2, 0.5, alpha[i]))
    
    #=============================================================================
    # THE FOLLOWING METHODS ARE NOT TESTED YET AND ALMOST NO DOCUMENTATION
    #=============================================================================    
    def euler_newton(self, q, dq, ddq):
        """ Euler-Netwon recursive algorithm from book Sicilliano """
        ndof = self.ndof
        lq = [len(vec) for vec in [q, dq, ddq]]
        if not np.all(np.array(lq) == ndof):
            raise ValueError("Input vectors must have same length, not: " + str(lq))
        
        # initialize vectors for solution (base at index 0)
        # i = 0 -> base
        # i = ndof+1 - > last link
        # assume base is not moving TODO
        W  = np.zeros(ndof + 1) # link angular velocity
        dW = np.zeros(ndof + 1) # link angular acceleration
        A  = np.zeros((ndof + 1, 2)) # frame acceleration
        Ac = np.zeros((ndof + 1, 2)) # link cg acceleration
        
        # forward recursion for speed and acceleration
        for k in range(1, ndof+1):
            i = k-1 # link index in robot parameters
            W[k], dW[k], A[k], Ac[k] = self._fw_prop(i, q[i], dq[i], ddq[i],
                                                     W[k-1], dW[k-1], A[k-1])
        
        # initialize some more vectors (OTHER INDEX DEFINITION than W, dW, ...)
        # i = ndof+1 -> end effector
        # i = 0 - > first link
        # assume no end effector force or torque
        F   = np.zeros((ndof + 1, 2))
        M   = np.zeros(ndof + 1)
        tau = np.zeros(ndof) # joint force or torque
        
        # backward recursion for forces and torques
        for k in np.flip(np.arange(ndof), 0):
            i = k # link index in robot parameters
            F[k], M[k], tau[k] = self._bw_prop(i, q[i],
                                             F[k+1], M[k+1],
                                             dW[k+1], Ac[k+1])
        return tau
    
    def _fw_prop(self, ib, qb, dqb, ddqb, wa, dwa, aa):
        """ Forward propagation of speed and velocity from link i to i+1 """
        
        cb = self.c[ib]
        if self.jt[ib] == 'r':
            db    = self.d[ib]
            arel  = [-(wa + dqb)**2 * db, (dwa + ddqb)*db]
            acrel = [-(wa + dqb)**2 * cb, (dwa + ddqb)*cb]
            wb    = wa  + dqb
            dwb   = dwa + ddqb
            R = self._R_link(ib, qb).T
        elif self.jt[ib] == 'p':
            db    = qb
            arel  = [-wa**2 * db + ddqb, dwa*db + 2*wa*dqb]
            acrel = [-wa**2 * cb + ddqb, dwa*cb + 2*wa*dqb]
            wb    = wa
            dwb   = dwa
            R = self._R_link(ib, self.a[ib]).T
        else:
            raise ValueError("wrong joint type: " + self.jt[ib])
        
        ab = np.dot(R, aa) + np.array(arel)
        ac = np.dot(R, aa) + np.array(acrel)
        return wb, dwb, ab, ac
    
    def _bw_prop(self, ia, qb, Fb, Mb, dwa, aca):
        """ Backward force and torque propagation from i+1 to i """
        
        if self.jt[ia] == 'r':
            # transforme Fb to frame of link a (given in frame b)
            R = self._R_link(ia+1, qb)
            Fbt = np.dot(R, Fb)
            # calculate Fa and Ma
            Fa = Fbt + self.m[ia] * aca
            Ma = Mb  + self.c[ia] * Fa[1]
            Ma += (self.d[ia] - self.c[ia]) * Fb[1]
            Ma += self.I[ia] * dwa
            # joint torque
            taua = Ma
        elif self.jt[ia] == 'p':
            # transforme Fb to frame of link a (given in frame b)
            R = self._R_link(ia+1, self.a[ia+1])
            Fbt = np.dot(R, Fb)
            # calculate Fa and Ma
            Fa = Fbt + self.m[ia] * aca
            Ma = Mb  + self.c[ia] * Fa[1]
            Ma += (qb - self.c[ia]) * Fb[1]
            Ma += self.I[ia] * dwa
            # joint force
            taua = Fa[0]
        else:
            raise ValueError("wrong joint type: " + self.jt[ia])
        
        return Fa, Ma, taua
    
    def _R_link(self, i, qi):
        """ rotation matrix of link i relative to link i-1
        
        Add check for if i == 0?
        
        Parameters
        ----------
        i : int
            link index, numbered from 0 to ndof
        qi : float
            joint value for joint i, numered from 0 to ndof, correspond
            to the link with the same index.
        
        Returns
        -------
        numpy.ndarray
            Array with shape (2, 2), a 2D rotation matrix representing link i
            relative to link i-1
        """
        if (i+1) >= self.ndof:
            # frame i is the end effector frame
            # end effector pose not implemented
            return np.eye(2)
        else:
            if self.jt[i] == 'r':
                a = qi
            elif self.jt[i] == 'p':
                a = self.a[i]
            else:
                raise ValueError("wrong joint type: " + self.jt[i])
            
            ca = np.cos(a)
            sa = np.sin(a)
            R = np.array([[ca, -sa], [sa, ca]])
            return R

class Robot_3R(Robot):
    """ Wrapper class for base robot class for 3R robot.
    
    This class adds analytical inverse kinematics for this specific robot.
    """
    def __init__(self, link_length):
        """ Simplified constructor for this 3R robot
        
        Parameters
        ----------
        link_length : list or np.array of floats
            The lengths for the three links.
        """
        if not (len(link_length) == 3):
            raise ValueError("This robot has 3 links, not: " + str(len(link_length)))
        super().__init__(['r', 'r', 'r'], link_length, [0, 0, 0])
    
    def ik(self, p, tol=1e-12):
        """ Analytic inverse kinematics for 3 link robot
        
        Joint limits [-pi, pi] are implied by the way the solution is
        calculated. This should be fixed.  
        
        Parameters
        ----------
        p : list or np.ndarray of floats
            End-effector pose (x, y, angle)
        tol : float
            Tolerance when a result is assumed zero. (default = 1e-12)
        
        Returns
        -------
        dict
            A dictionary with a key 'success' reporting True if the pose is
            reachable and a key 'q' reporting the different joint solutions
            as a list of numpy arrays.
            If 'success' is False, a key 'info' containts extra info.
        
        Notes
        -----
        Joint limits have to be added.
        """
        # tol = tolerance needed for boundary cases
        # base transform only translation for now TODO
        p = np.array(p)
        p[:2] = p[:2] - self.base[:2]
        
        # define variables for readability
        l1, l2, l3 = (self.d[0], self.d[1], self.d[2])
        x, y, phi = (p[0], p[1], p[2])
        
        # initialize output
        q_up = np.zeros(3)
        q_do = np.zeros(3)
        reachable = False
        
        # start calculations
        if (l1 + l2 + l3) >= np.sqrt(x**2 + y**2):
            # coordinates of end point second link (w)
            pwx = x - l3 * np.cos(phi)
            pwy = y - l3 * np.sin(phi)
            rws = pwx**2 + pwy**2 # squared distance to second joint
            if (l1 + l2) >= np.sqrt(rws):
                # calculate cosine q2
                c2 = (rws - l1**2 - l2**2) / (2*l1*l2)
                # c2 is already guaranteed to be < 1
                # but it can still be smaller than -1
                # for example points close the the robot base that are
                # not reachable
                if c2 > -1:
                    reachable = True
                    # if c2 exactly 1, it can be a little bit bigger at this point
                    # because of numerical error, then rescale
                    # TODO far from all edge cases are catched
                    if abs(c2 - 1) < tol:
                        c2 = np.sign(c2) * 1.0
                        s2 = 0.0
                        q_up[1] = np.arctan2(0, c2) # elbow up
                        q_do[1] = -np.arctan2(0, c2) # elbow down
                    else:
                        s2 = np.sqrt(1 - c2**2)
                        q_up[1] = np.arctan2(s2, c2) # elbow up
                        q_do[1] = np.arctan2(-s2, c2) # elbow down
                    # calculate q1
                    temp = (l1 + l2 * c2)
                    s1_up = (temp * pwy - l2 * s2 * pwx) / rws
                    c1_up = (temp * pwx + l2 * s2 * pwy) / rws
                    s1_do = (temp * pwy + l2 * s2 * pwx) / rws
                    c1_do = (temp * pwx - l2 * s2 * pwy) / rws
                    q_up[0] = np.arctan2(s1_up, c1_up)
                    q_do[0] = np.arctan2(s1_do, c1_do)
                    # finally q3
                    q_up[2] = phi - q_up[0] - q_up[1]
                    q_do[2] = phi - q_do[0] - q_do[1]

        if reachable:
            # check if we have to identical solutions
            if np.allclose(q_up, q_do):
                return {'success': True, 'q': [q_up]}
            else:
                return {'success': True, 'q': [q_up, q_do]}
        else:
            return {'success': False, 'info': "unreachable"}

class Robot_2P(Robot):
    """ Wrapper class for base robot class for 2P robot.
    
    A simple cartesian robot with inverse kinematics.
    First joint along x-axis, second joint along y-axis.
    """
    def __init__(self, link_length):
        """ Simplified constructor for this 2P robot
        
        Parameters
        ----------
        link_length : list or np.array of floats
            The lengths for the two links.
        """
        if not (len(link_length) == 2):
            raise ValueError("This robot has 2 links, not: " + str(len(link_length)))
        super().__init__(['p', 'p'], link_length, [0, np.pi / 2])
    
    def ik(self, p, tol=1e-12):
        """ Inverse kinematics for 2R robot, ignoring orientation.
        
        Joint limits not implemented yet.
        
        Parameters
        ----------
        p : list or np.ndarray of floats
            End-effector pose (x, y, angle)
            Angle (orientation) not used but included for consistency.
        tol : float
            Tolerance when a result is assumed zero. (default = 1e-12)
            Not used but included for consistency.
        
        Returns
        -------
        dict
            A dictionary with a key 'success' reporting True if the pose is
            reachable and a key 'q' reporting the different joint solutions
            as a list of numpy arrays.
            If 'success' is False, a key 'info' containts extra info.
        """
        return {'success': True, 'q': [np.array([p[0], p[1]])]}

class Robot_2P3R(Robot):
    """ A 3R robot mounted on a 2P cartesian robot
    
    This is the first robot with kinematic redundancy!
    
    Attributes
    ----------
    ik_sample : int
            How many samples should be taken in the range of the first two
            joints. Resulting in n_sample * n_sample times the normal number
            of solutions for a 3R robot.
    """
    def __init__(self, link_length, ik_samples=[5, 5]):
        """ Simplified constructor for this 2P3R robot
        
        Add a helper robot to self.sub_robot to reuse the inverse kinematics
        implementation of the 3R robot.
        
        Parameters
        ----------
        link_length : list or np.array of floats
            The lengths for the two links.
        ik_samples : list of int
          Number of samples that should be taken for the two fixed joints.
        """
        if not (len(link_length) == 5):
            raise ValueError("This robot has 5 links, not: " + str(len(link_length)))
        super().__init__(['p', 'p', 'r', 'r', 'r'],
                         link_length,
                         [np.pi / 2, -np.pi / 2, 0, 0, 0])
        # create 3R robot for inverse kinematics
        self.sub_robot = Robot_3R(link_length[2:])
        self.ik_samples = ik_samples

    def ik(self, p):
        """ Discretised / sampled inverse kinematics
        
        This robots has redundance (ndof = 5) compared to the task (3) and
        therefore two joints are sampled in a range to return a sampled
        subset of the infinite solutions that exists for the given pose.
        
        Parameters
        ----------
        p : list or np.ndarray of floats
            End-effector pose (x, y, angle)
        
        Returns
        -------
        dict
            A dictionary with a key 'success' reporting True if the pose is
            reachable and a key 'q' reporting the different joint solutions
            as a list of numpy arrays.
            If 'success' is False, a key 'info' containts extra info.
        """
        if hasattr(self, 'jl'):
            jl1, jl2 = self.jl[0], self.jl[1]
        else:
          # use the middle of the interval as nominal value
          jl1, jl2 = (0, 1.5), (0, 1.5)
        # nominal value in the middle of the limits
        n1 = (jl1[0] + jl1[1]) / 2
        n2 = (jl2[0] + jl2[1]) / 2
        
        # create sampled values for fixed joints and put them in a grid
        q1 = TolerancedNumber(n1, jl1[0], jl1[1], samples=self.ik_samples[0])
        q2 = TolerancedNumber(n2, jl2[0], jl2[1], samples=self.ik_samples[1])
        grid = np.meshgrid(q1.range, q2.range)
        grid = [ grid[i].flatten() for i in range(2) ]
        grid = np.array(grid).T
        
        q_sol = []
        for qf in grid:
            s = self.ik_fixed_joints(p, q_fixed=qf)
            if s['success']:
                for qi in s['q']:
                    q_sol.append(qi)
        if len(q_sol) > 0:
            return {'success': True, 'q': q_sol}
        else:
            return {'success' : False, 'info': "unreachable"}
    
    def ik_fixed_joints(self, p, q_fixed = [0, 0]):
        """ wrapper function to solve the ik for the last three joints.
        
        The joint solution of the 3R robot is extended with the joint values
        for the fixed joints.
        
        Parameters
        ----------
        p : list or np.ndarray of floats
            End-effector pose (x, y, angle) 
        dict
            A dictionary with a key 'success' reporting True if the pose is
            reachable and a key 'q' reporting the different joint solutions
            as a list of numpy arrays.
            If 'success' is False, a key 'info' containts extra info.
        """
        q_base = [0, 0, 0]
        q_base[0] = q_fixed[1]
        q_base[1] = q_fixed[0]
        self.sub_robot.base = q_base
        sub_sol = self.sub_robot.ik(p)
        if sub_sol['success']:
            # add fixed joints to solution
            q_sol = []
            for qi in sub_sol['q']:
                q_sol.append([*q_fixed, *qi])
            sub_sol['q'] = q_sol
        return sub_sol


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes and function related to solving optimisation problems related
to robot path following.
"""

import numpy as np
from scipy.optimize import fmin_slsqp

#=============================================================================
# Use general minimize interface and implement object oriented
#=============================================================================

class Solver():
    """ Setup and solve an optimal robot motion planning problem
    
    No dual variables implemented at the moment. Look in branch old-master
    optimize.py to see how this can be added.
    """
    def __init__(self, robot, path, scene, w=None):
        self.robot = robot
        self.path = path
        self.scene = scene
        
        # set default weights
        if w == None:
            self.w = {'joint_motion': 1.0,'path_error': 0.0, 'torque': 0.0}
        else:
            self.w = w
            
        self.bnds = []
        self.cons = []
    
    def create_objective(self):
        ndof = self.robot.ndof
            
        # put weights in list
        wb = [w[key] for key in self.w]
        
        # if torque objectives, check dynamics
        if wb[2] > 0:
          if not hasattr(robot, 'c'):
            msg = "Specify robot dynamic parameter s"
            msg += "using the Robot.set_link_inertia() method"
            raise ValueError(msg)
        
        # setup the desired objective
        if wb[0] > 0 and wb[1] == 0 and wb[2] == 0:
            def obj(x):
                n_path, qp = reshape_path_vector(x, n_dof=ndof)
                return joint_motion_obj(qp)
        elif wb[1] > 0 and wb[2] == 0:
            def obj(x):
                n_path, qp = reshape_path_vector(x, n_dof=ndof)
                return path_error_obj(qp, robot, self.path)
        elif wb[2] > 0:
            def obj(x):
                n_path, qp = reshape_path_vector(x, n_dof=ndof)
                return torque_obj(qp, robot)
        elif wb[0] > 0 and wb[1] > 0 and wb[2] == 0:
            def obj(x):
                n_path, qp = reshape_path_vector(x, n_dof=ndof)
                res =  wb[0] * joint_motion_obj(qp)
                res += wb[1] * path_error_obj(qp, self.robot, self.path)
                return res
        else:
            raise ValueError("This type of objective is not implemented yet.")
        
        return obj
    
    def add_path_constraints(self):
        def path_con(x):
            n_path, qp = reshape_path_vector(x, n_dof=self.robot.ndof)
            return path_ieq_con(qp, self.robot, self.path)
        self.cons.append({'type': 'ineq', 'fun': path_con})
    
    def add_collision_constaints(self):
        def cc_con(x):
            n_path, qp = reshape_path_vector(x, n_dof=self.robot.ndof)
            return collision_ieq_con(qp, self.robot, self.scene)
        self.cons.append({'type': 'ineq', 'fun': cc_con})
    
    def run(self, q_init):
        obj  = self.create_objective()
        return minimize(obj, q_init, method='SLSQP', 
                        constraints=self.cons,
                        bounds = self.bnds)


#=============================================================================
# Main functions to run the problem
#=============================================================================
def get_optimal_trajectory(robot, path, q_path_init,
                           check_collision=False, scene=None,
                           w={'joint_motion': 1.0,
                               'path_error'  : 0.0,
                               'torque'      : 0.0}):
    """ Formulate and solve optimal path following problem
    
    Parameters
    ----------
    robot : ppr.robot.Robot
    path : list of ppr.path.TrajectoryPt
    q_path_init : list of numpy.ndarray
    w : dict
    
    Returns
    -------
    tuple of numpy.ndarray
        Joint position, velocity and acceleration
    """
    # input validation
    if check_collision:
        if scene == None:
            raise ValueError("scene is needed for collision checking")
    
    n_path = len(path)
    q_init = np.array(q_path_init).flatten()

    # setup the correct objective
    wb = [w[key] for key in w]
    # if torque objectives, check dynamics
    if wb[2] > 0:
      if not hasattr(robot, 'c'):
        msg = "Specify robot dynamic parameter s"
        msg += "using the Robot.set_link_inertia() method"
        raise ValueError(msg)
    if wb[0] > 0 and wb[1] == 0 and wb[2] == 0:
        def obj(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            return joint_motion_obj(qp)
    elif wb[1] > 0 and wb[2] == 0:
        def obj(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            return path_error_obj(qp, robot, path)
    elif wb[2] > 0:
        def obj(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            return torque_obj(qp, robot)
    elif wb[0] > 0 and wb[1] > 0 and wb[2] == 0:
        def obj(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            res =  wb[0] * joint_motion_obj(qp)
            res += wb[1] * path_error_obj(qp, robot, path)
            return res
    else:
        raise ValueError("This type of objective is not implemented yet.")

    # setup inequality constraints
    if check_collision:
        def ieq_con(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            pc = path_ieq_con(qp, robot, path)
            cc = collision_ieq_con(qp, robot, scene)
            return np.hstack((pc, cc))
    else:
        def ieq_con(x):
            n_path, qp = reshape_path_vector(x, n_dof=robot.ndof)
            return path_ieq_con(qp, robot, path)

    # Solve problem
    sol = fmin_slsqp(obj,
                     q_init,
                     f_ieqcons=ieq_con)

    n_path, qp_sol = reshape_path_vector(sol, n_dof=robot.ndof)
    dqs, ddqs = q_derivatives(qp_sol)
    return qp_sol, dqs, ddqs

#=============================================================================
# Objectives
# function structure : xxxx_obj(opt_var, params ...)
    # xxxxx   : name, the thing that is minimized
    # opt_var : the optimizationsv variables
    # parmas  : parameters, such as robot, path, coefficients, ....
# returns
    # a a scalar that has to be minimized
#=============================================================================

def path_error_obj(q_path, robot, path):
    con = []
    for i, tp in enumerate(path):
        pp = tp.p_nominal
        pfk = robot.fk(q_path[i])
        con.append(np.sum((pp - pfk)**2))
    con = np.array(con).flatten()
    return np.sum(con**2)

def joint_motion_obj(q_path):
    dqp = np.diff(q_path, axis=0)

    return np.sum(np.abs(dqp))

def torque_obj(q_path, robot):
    n_path, ndof = q_path.shape
    dqp, ddqp = q_derivatives(q_path)
    tau = np.zeros((n_path, robot.ndof))
    for i in range(n_path):
        tau[i, :] = robot.euler_newton(q_path[i], dqp[i], ddqp[i])
    return np.sum(tau**2)

#=============================================================================
# Constraints
    # function structure : xxxx_eq_con(opt_var, params ...)
    #                    : xxxx_ieq_con(opt, var, params ...)
    # xxxxx   : name, the thing that causes the constraints
    # opt_var : the optimizationsv variables
    # parmas  : parameters, such as robot, path, coefficients, ....
# returns
    # equality constraints: a numpy array, each element has to be zero
    # inequality constraints: a numpy array, each elment is positive when
    # the constraints is satisfied
#=============================================================================

def path_ieq_con(q_path, robot, path, tol=1e-6):
    con = []
    for i, tp in enumerate(path):
        has_tol = tp.hasTolerance
        pfk = robot.fk(q_path[i])
        for i in range(tp.dim):
            if has_tol[i]:
                con.append(-pfk[i] + tp.p[i].u)
                con.append( pfk[i] - tp.p[i].l)
            else:
                con.append(-pfk[i] + tp.p[i] + tol)
                con.append( pfk[i] - tp.p[i] - tol)
    return np.array(con)

def collision_ieq_con(q_path, robot, scene):
    con = []
    for qi in q_path:
        robot_shapes = robot.get_shapes(qi)
        for s1 in robot_shapes:
            for s2 in scene:
                con.append(s1.distance(s2))
    
    return np.array(con)

#=============================================================================
# Utility functions
#=============================================================================

def reshape_path_vector(v_flat, n_dof=3):
    """ Convert flat vector to n_path x n_dof path array

    n_path is the number of discrete path points.
    This reshaping if often needed to formulate optimization problems.
    The standard optimization variable shape in scipy.optimize is a
    1D vector.

    Parameters
    ----------
    v_flat : ndarray
        Vector of length n_path*ndof containing the joint position vector for
        every path point [q1, q2, ...] with q1 of length ndof.
    ndof : int
        Degrees of freedom of robot.

    Returns
    -------
    tuple of int and numpy.ndarray
        Tuple with first element the number of path poitns and the second
        element an array of shape (n_path, ndof) every row containt the joint
        positions for a single path point.
    
    Examples
    --------
    >>> a = np.array([1, 2, 3, 4, 5, 6])
    >>> N, a_new = reshape_path_vector(a)
    >>> N
    2
    >>> a_new
    array([[1, 2, 3],
           [4, 5, 6]])
    """
    n_path = int(len(v_flat) / n_dof)
    return n_path, v_flat.reshape(n_path, n_dof)

def q_derivatives(q, dt=0.1):
    """ Calculate joint speed and acceleration

    Based on a given path, caculate speed and acceleration using the
    numpy.gradient function. A constant sample time dt is assumed.

    Parameters
    ----------
    q : ndarray
        Array of shape (n_path, ndof) every row containt the joint positions for
        a single path point.
    dt : float
        Sample time for joint position path

    Returns
    -------
    tuple
        (dq, ddq) the joint speed and acceleration along the path.
        Arrays with the same shape as the input path array q.
    """
    dq = np.gradient(q, dt, axis=0)
    ddq = np.gradient(dq, dt, axis=0)
    return dq, ddq




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_almost_equal
from ppr.util import cp, VirtualArray, VirtualJointSolutions
from ppr.path import TolerancedNumber, TrajectoryPt
from ppr.robot import Robot_2P3R
from ppr.geometry import Rectangle

def test_cp():
  l = [2, 3, 6]
  a = cp(l)
  d = 2*3*6
  assert_almost_equal(a, d)

class TestVirtualArray():
  def test_indexing(self):
    ranges = [np.array([1, 2]),
              np.array([3]),
              np.array([5, 6, 7])]
    va = VirtualArray(ranges)
    ac = [va[i] for i in range(va.total_size)]
    de = [[1, 3, 5],[2, 3, 5],
          [1, 3, 6],[2, 3, 6],
          [1, 3, 7],[2, 3, 7]]
    assert_almost_equal(ac, de)

class TestVirtualJointSolutions():
  # SLOW test (0.2 s)
  def test_all(self):
    # create trajectory point
    x = TolerancedNumber(2, 1, 3, samples=3)
    y = TolerancedNumber(2.5, 2, 3, samples=4)
    tp = TrajectoryPt([x, y, 0])
    
    # create robot and collision scene
    robot1 = Robot_2P3R([1, 1, 2, 1.5, 0.5])
    sc1 = [Rectangle(2, 1, 0.5, 0.5, 0)]
    
    # create the object
    vjs = VirtualJointSolutions(robot1, tp, sc1)
    
    # try to index the whole range
    for i in range(vjs.size): sol = vjs[i]
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from scipy.linalg import norm

def rotation(angle):
    """ Create 2x2 rotation matrix from angle
    
    Parameters
    ----------
    anlge : float
    
    Returns
    -------
    numpy.ndarray
        Rotation matrix as a numpy array
    
    Examples
    --------
    >>> rotation(0.0)
    array([[ 1., -0.],
           [ 0.,  1.]])
    >>> rotation(np.pi / 2)
    array([[  6.12323400e-17,  -1.00000000e+00],
           [  1.00000000e+00,   6.12323400e-17]])
    """
    return np.array([[np.cos(angle),  -np.sin(angle)],
                      [np.sin(angle),  np.cos(angle)]])

class Rectangle:
    """ Rectangle plotting, handling and collision detection
    
    Attributes
    ----------
    p : numpy.ndarray
        The four corner points of the rectangle, counter-clockwise, starting
        with the point at the (x, y) position. The array has shape (4, 2).
    
    Examples
    --------
    >>> rect = Rectangle(0.3, 0.4, 1.5, -1.6, 2.7)
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> rect.plot(ax)
    """
    def __init__(self, x, y, dx, dy, angle):
        """ Create a rectangle object
        
        Parameters
        ----------
        x : float
            x-position of rectangle, one of the four corner points.
        y : float
            y-position of rectangle, one of the four corner points.
        dx : float
            Width before rotation.
        dy : float
            Height before roation.
        angle : float
            Angle between x-axis and bottom side of rectangle.
        """
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.a = angle
        self.R = rotation(angle)
        self.p = self.get_vertices()

    def get_vertices(self):
        """ Get 4 corner points of rectangles
        
        Returns
        -------
        numpy.ndarray
            Array with shape (4, 2) containing the x and y coordinates of
            the 4 corners of the rectangle.
        
        Examples
        --------
        >>> rec1 = Rectangle(0, 0, 1, 1, 0)
        >>> rec1.get_vertices()
        array([[ 0.,  0.],
               [ 1.,  0.],
               [ 1.,  1.],
               [ 0.,  1.]])
        """
        p = np.zeros((4, 2))
        p[0, 0] = self.x
        p[0, 1] = self.y
        p[1, :] = np.dot(self.R, [self.dx, 0 ]) + p[0, :]
        p[2, :] = np.dot(self.R, [self.dx, self.dy]) + p[0, :]
        p[3, :] = np.dot(self.R, [0 ,      self.dy]) + p[0, :]
        return p

    def get_normals(self):
        """ Get normals on the 4 sides
        
        Get 4 unit vectors pointing away from the four sides.
        Start with [0, -1] and rotate it by the angle of the rectangle.
        Then rotate it 3 times with a angle pi/4.
        
        Returns
        -------
        numpy.ndarray
            Array with shape (4, 2) containing 4 unit vectors representing
            normals on 4 rectangle sides.
        
        Examples
        --------
        >>> rec1 = Rectangle(0, 0, 1, 1, 0)
        >>> rec1.get_normals()
        array([[  0.00000000e+00,  -1.00000000e+00],
               [  1.00000000e+00,  -6.12323400e-17],
               [  1.22464680e-16,   1.00000000e+00],
               [ -1.00000000e+00,   1.83697020e-16]])
        """
        p = self.p
        n = np.zeros((4, 2))
        n[0, :] = np.dot(self.R, np.array([0.0, -1.0]))
        Rtemp = rotation(np.pi/2)
        n[1, :] = np.dot(Rtemp, n[0, :])
        n[2, :] = np.dot(Rtemp, n[1, :])
        n[3, :] = np.dot(Rtemp, n[2, :])
        return n
    
    def project(self, axis):
        """ Project all points of rectangle on an axis given as unit vector
        
        Use the dot product between the vector from the origin to a corner
        point and the given axis unit vector.
        
        Parameters
        ----------
        axis : numpy.ndarray of floats
            Unit vector representing the axis on wich to project the points.
        
        Returns
        -------
        numpy.ndarray
            Array with shape (4,) containing the projected points with the
            origin as reference.
        
        Examples
        --------
        
        Project simple square on x-axis.
        
        >>> rec1 = Rectangle(0, 0, 1, 1, 0)
        >>> rec1.project(np.array([1, 0]))
        array([ 0.,  1.,  1.,  0.])
        
        Project the same square on diagonal axis.
        
        >>> rec1.project(np.array([0.70710678, 0.70710678]))
        array([ 0.        ,  0.70710678,  1.41421356,  0.70710678])
        """
        return np.dot(self.p, axis)
    
    def is_in_collision(self, rect2, tol=1e-9):
        """ Check if it collides with another rectangle.
        
        Use the separating axis theorem.
        Project both rectangles along all 8 normals and check overlap.
        
        Parameters
        ----------
        rect2 : ppr.geometry.Rectangle
            The other rectangle to check collision with.
        
        Returns
        -------
        bool
            False if no collision (as soon as a separating axis is found),
            True if in collision.
        
        Examples
        --------
        >>> rec1 = Rectangle(0, 0, 1, 1, 0)
        >>> rec2 = Rectangle(0.5, 0, 1, 1, 0.1)
        >>> rec3 = Rectangle(1.5, 0.5, 1, 2, -0.2)
        >>> rec1.is_in_collision(rec2)
        True
        >>> rec1.is_in_collision(rec3)
        False
        """
        n1 = self.get_normals()
        n2 = rect2.get_normals()
        n_all = np.vstack((n1, n2))
        # assume collision until proven otherwise
        col = True
        i = 0
        while col and i < 8:
            pr1 = self.project(n_all[i])
            pr2 = rect2.project(n_all[i])
            if (( max(pr1) + tol < min(pr2) ) or ( min(pr1) > max(pr2) + tol )):
                col = False
            i += 1

        return col
    
    def distance(self, rect2, tol=1e-9):
        """ Check if it collides with another rectangle.
        
        Use the separating axis theorem.
        Project both rectangles along all 8 normals and check overlap.
        
        Parameters
        ----------
        rect2 : ppr.geometry.Rectangle
            The other rectangle to calculate the separating distance with.
        
        Returns
        -------
        float
            Distance between the two rectangles. Negative penetration depth
            if in collision
        
        Examples
        --------
        >>> rec1 = Rectangle(0, 0, 1, 1, 0)
        >>> rec2 = Rectangle(0.5, 0, 1, 1, 0.1)
        >>> rec3 = Rectangle(1.5, 0.5, 1, 2, -0.2)
        >>> rec1.distance(rec2)
        -0.59733549928584095
        >>> rec1.distance(rec3)
        0.5
        """
        n1 = self.get_normals()
        n2 = rect2.get_normals()
        n_all = np.vstack((n1, n2))
        # assume collision until proven otherwise
        col = True
        dist = 0.0
        i = 0
        while col and i < 8:
            pr1 = self.project(n_all[i])
            pr2 = rect2.project(n_all[i])
            d1 = min(pr2) - max(pr1)
            d2 = min(pr1) - max(pr2)
            if (d1 > tol or d2 > tol):
                col = False
            dist = max(d1, d2)
            i += 1
            
        return dist
    
    def get_matrix_form(self):
        """ Get the matrix representation of the rectanlge
        
        A rectangle is described by four linear inequalities.
        If whe write these inequalities in matrix form we get
        A * x <= b
        
        x is the vector of all possible x and y coordinates [x; y]
        
        Returns
        -------
        A : numpy.ndarray of floats
            Array with shape (4, 2), every rows contains the coefficient of
            on of the inequalities.
        b : numpy.ndarray fo floats
            Vector of length 4 containing the right hand side of the
            inequalities.
            
        Examples
        --------
        
        Show the inequalities for a trivial case.
        For a 1x2 rectange at the origin whe whould expect:
        -y <= 0
        x <= 1
        y <= 2
        -x <= 0
        
        >>> rec1 = Rectangle(0, 0, 1, 2, 0)
        >>> A1, b1 = rec1.get_matrix_form()
        >>> A1
        array([[  0.00000000e+00,  -1.00000000e+00],
               [  1.00000000e+00,  -6.12323400e-17],
               [  1.22464680e-16,   1.00000000e+00],
               [ -1.00000000e+00,   1.83697020e-16]])
        >>> b1
        array([  0.00000000e+00,   1.00000000e+00,   2.00000000e+00,
                 3.67394040e-16])
        """
        A = self.get_normals()
        # row wise dot product
        b = np.sum(A * self.get_vertices(), axis=1)
        return A, b

    def plot(self, ax, *arg, **karg):
        p = self.p
        p = np.vstack((p, p[0]))
        ax.plot(p[:, 0], p[:, 1], *arg, **karg)



from setuptools import setup, find_packages

setup(
    name = 'ppr',
    version = '0.1.1',
    packages = find_packages(),
    description = 'Planar Python Robotics',
    long_description=('Software tool to experiment with 2D motion planning problems' +
    'for robot manipulators.'),
    author = 'Jeroen De Maeyer',
    author_email = 'jeroen.demaeyer@kuleuven.be',
    url = 'https://u0100037.pages.mech.kuleuven.be/planar_python_robotics/',
    download_url = 'https://gitlab.mech.kuleuven.be/u0100037/planar_python_robotics/raw/master/dist/ppr-0.1.1.tar.gz',
    keywords = ['robotics', 'motion planning'],
    classifiers = [],
    install_requires=['scipy', 'matplotlib'],
    python_requires='>=3',
)



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_graph', [dirname(__file__)])
        except ImportError:
            import _graph
            return _graph
        if fp is not None:
            try:
                _mod = imp.load_module('_graph', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _graph = swig_import_helper()
    del swig_import_helper
else:
    import _graph
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0


class Graph(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Graph, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Graph, name)
    __repr__ = _swig_repr

    def add_data_column(self, mat):
        return _graph.Graph_add_data_column(self, mat)

    def print_graph_data(self):
        return _graph.Graph_print_graph_data(self)

    def init_dijkstra(self):
        return _graph.Graph_init_dijkstra(self)

    def run_dijkstra(self):
        return _graph.Graph_run_dijkstra(self)

    def get_path(self, vec):
        return _graph.Graph_get_path(self, vec)

    def print_path(self):
        return _graph.Graph_print_path(self)

    def print_graph(self):
        return _graph.Graph_print_graph(self)

    def set_graph_data(self, data):
        return _graph.Graph_set_graph_data(self, data)

    def __init__(self):
        this = _graph.new_Graph()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _graph.delete_Graph
    __del__ = lambda self: None
Graph_swigregister = _graph.Graph_swigregister
Graph_swigregister(Graph)
cvar = _graph.cvar
INF = cvar.INF

# This file is compatible with both classic and new-style classes.


